from pearlib.get_lines_from_layer_output import*
from pearlib.get_color_for_layeridx import *
from pearlib.network import *


def get_all_lines(network):
    """Stores and returns an array containing all the boundary lines for all layers.
       -- network: network object (containing the outputs of the layers in itself)
        Output: If we had n layers: n long list. The list contains lists of lines.
            The list ith element is a list, containing the boundary lines for the ith layer (if the ith layer contained
            m_i neurons, the list is m_i long.)
    """
    # Get the outputs of the layers in the network
    all_output = network.get_all_output()
    all_lines = []

    # For each layer
    for i in range(len(all_output)):
        # Get the lines from it
        lines = get_lines_from_layer_output(network, i, get_color_for_layeridx(i))
        # Append them to the all_lines array
        all_lines.append(list(lines))

    return all_lines



from pearlib.get_all_lines import *
from pearlib.create_image_from_lines import *


def create_image(network, layer_indices, res = None, img = None, plot_output = False):
    """Generates the image with the boundary lines.
        -- network: The Network object representing the network
        -- layer_indices: indicates which layers we would like to draw
        -- res: resolution of the plane ([n_col, n_row])
        -- img: optional. If given, the lines will be plotted to that image. If given, its resolution is used
                for the output image
        -- plot_output: optional.
            If true, the background of the image will be the output of the neuron in the last layer
    """

    # If neither img nor res is given -> error, we do not know the resolution
    if img is None and res is None:
        raise ValueError("No resolution specified")

    # If img is given, we set res for the resolution of the image
    if img is not None:
        res = [img.shape[1], img.shape[0]]

    # Calculate the output of the network for the specified resolution
    network.compute_all_output(res)
    # Get the boundary lines
    network.compute_all_lines()
    all_the_lines = network.get_lines()

    # Return the created image
    return create_image_from_lines(all_the_lines, layer_indices, res = res, img = img, network = network, plot_output = plot_output)



from tensorflow.python.training import *
from tensorflow import errors
from tensorflow.python.framework.errors_impl import NotFoundError
from tensorflow import Graph
from tensorflow import get_default_graph
from tensorflow import Session


def read_tensorflow_file(fname):
    """Read a TensorFlow file given with its filename.
            Return the parameters (weight matrix, bias) of the network.
    """
    new_graph = Graph()
    theta = []
    tmpW = []
    tmpb = []

    with Session(graph=new_graph) as sess:
        # Handles case where there is no .meta file
        try:
            saver.import_meta_graph(fname + '.meta')
        except OSError:
            raise OSError("The .META file could not be found!")

        graph = get_default_graph()

        # Handles the case where data is corrupted or missing
        try:
            saver.Saver().restore(sess, fname)
        except errors.DataLossError:
            raise errors.DataLossError(None, None,
                                       "The data seems to be corrupted and the network could not be reconstructed!")
        except NotFoundError:
            raise NotFoundError(None, None, "A file (.INDEX or .DATA) could not be found!")

        theta = []
        tmpW = [0] * int(len(graph.get_collection_ref('trainable_variables')) / 2)
        tmpb = [0] * int(len(graph.get_collection_ref('trainable_variables')) / 2)

        for x in graph.get_collection_ref('trainable_variables'):
            if x.name[0] == 'W' or x.name[0] == 'b':
                if x.name[0] == 'W':
                    tmpW[int(x.name[1 : -2]) - 1] = x.eval(sess)
                else:
                    tmpb[int(x.name[1 : -2]) - 1] = x.eval(sess)

    if len(tmpW) != len(tmpb):
        raise errors.DataLossError(None, None,
                                   "The data is corrupted, or the naming was improper!")

    for x in range(len(tmpW)):
        theta += [(tmpW[x], tmpb[x])]

    # Checking that an actual network was loaded in, and not an empty one
    empty = True
    error = []
    for i in range(0, len(theta)):
        error.append((0, 0))
    try:
        error == theta
    except Exception:
        empty = False

    if empty == True:
        raise ValueError(
            "The network that was loaded in is empty! The input network you used probably does not follow our guidelines!")
    else:
        return theta


import numpy as np


class Layer:
    """Class for defining a layer with its weight matrix and bias vector"""
    W = []  # np.array!
    b = []  # np.array!
    n_neurons = 0
    n_input = 0
    color = [1, 1, 1]

    def __init__(self, _W, _b, _color):
        self.W = _W
        self.b = _b[:,np.newaxis]
        self.n_neurons = _W.shape[0]
        self.n_input = _W.shape[1]
        self.color = _color

    def calculate_layer_output_mat(self, *args):
        """Calculate the output of the layer for *args.
            args -- matrix containing n_input rows. Each column is treated as an input vector
            the result is the output of the layer for each input (for each column of args.)
        """
        # Convert the input to np.array
        input_vec = np.array(args)
        # Get rid of unwanted dimensions
        input_vec = input_vec.squeeze()
        # Check if the shape of the input is correct
        if input_vec.shape[0] != self.n_input:
            raise ValueError("The given input does not match the shape of the expected input of the network")

        # Calculate and return the output
        return np.matmul(self.W, input_vec) + self.b



import numpy as np
from pearlib.line import Line
from pearlib.network import *


def get_lines_from_layer_output(network, layer_idx, linecolor = [1, 1, 1]):
    """
    From the output, calculate the boundary lines.
    -- network: network object (containing the outputs of its layer)
    -- layer_idx: index, indicates that from which layer we want to get the lines
    -- linecolor: color of the line (color of the layer in which this line is)
    Return a list of lines, containing a line object for each neuron in the layer.
    If the layer contained m neuron, lines will be m long.
    """


    all_output = network.get_all_output()
    if layer_idx > len(all_output):
        raise ValueError("The layer index exceeds the size of the network")

    # Get the output of the layer
    h = all_output[layer_idx]
    # Get the sign for each element
    signed_h = np.sign(h)
    # Create an array to store if the sign changes at a given point of the input or not
    h_conv = np.zeros_like(signed_h)

    # For each neuron separately
    for k in range(0, signed_h.shape[2]):
        # Search for sign change in columns: If the sign changes from [i,j-1] to [i,j] than h_conv at [i,j]
        #   will contain 0, otherwise h[i,j] = 1. We do not examine where j=0. (first column)
        for i in range(0, signed_h.shape[0]):
            for j in range(0, signed_h.shape[1]):
                # first column: skip, here we can not have the boundary
                if j == 0:
                    h_conv[i, j, :] = 1
                # j > 0
                else:
                    # If the sign does not change
                    if signed_h[i, j, k] == signed_h[i, j-1, k]:
                        h_conv[i,j,k] = 1
                    # if the sign changes
                    else:
                        h_conv[i, j, k] = 0

        # Search for sing change in rows: almost the same as with columns, but we compare [i-1, j] with [i,j]
        # We only overwrite the elements if the sign differs between the rows, we do not set to '1'
        #   any element which was marked as 'changing sign' ones during examining the columns.
        for j in range(0, signed_h.shape[1]):
            for i in range(0, signed_h.shape[0]):
                # if it is the first row -> skip
                if i == 0:
                    h_conv[i, j, :] = 1
                # if the sign differs: set to 0
                else:
                    if signed_h[i, j, k] != signed_h[i-1, j, k]:
                        h_conv[i, j, k] = 0

    lines = []
    # For each neuron (channel) find the zero elements in h_conv. Convert them to Line objects,
    #   and append these Line objects ordered to the lines list
    for k in range(0, h_conv.shape[2]):
        lines.append(Line(np.array(np.where(h_conv[:, :, k] == 0)), _color = linecolor))

    return lines



from pearlib.reader import read_tensorflow_file
from pearlib.get_color_for_layeridx import get_color_for_layeridx
from pearlib.layer import Layer
from pearlib.calculate_all_output import calculate_all_output
from pearlib import get_all_lines # To avoid import conflicts with get_all_lines
import numpy as np


class Network:
    """Contains the structure of a neural network given as input."""

    def __init__(self, filename = None):
        """Create a new Network object.
        A Network object has five attributes:
        - a name "filename" (here, its full path without any extension)
        - parameters "theta" containing, for each layer, weight and bias
        - a list of layers "layers" containing Layer objects that correspond to a layer
        - all_outputs: a list containing the outputs of each layer - if it is calculated
        - all_lines: a list containing the boundary lines of each neuron - if it is calculated
        """
        self.__filename = filename
        self.__theta = read_tensorflow_file(self.__filename)
        self.__layers = self.create_layers(self.__theta)
        self.__all_output = []
        self.__all_lines = []

    def create_layers(self, theta):
        """Given the parameters "theta" of a network, extract the layers and build Layer objects from them."""
        layers = []
        for i in range(0, len(theta)):
            color = get_color_for_layeridx(i)
            W = np.transpose(theta[i][0])
            b = np.transpose(theta[i][1])
            layers.append(Layer(W, b, color))
        return layers

    def compute_all_output(self, res):
        """Computes the output of each neuron of the networkand stores it in __all_output"""
        self.__all_output = calculate_all_output(self.__layers, res)

    def compute_all_lines(self):
        """Computes the boundary lines for each neuron of the network and stores it in __all_lines"""
        # To avoid import errors, since * from network is imported in get_all_lines and there can be conflicts.
        self.__all_lines = get_all_lines.get_all_lines(self)

    def get_filename(self):
        return self.__filename

    def get_theta(self):
        return self.__theta

    def get_layers(self):
        return self.__layers

    def get_all_output(self):
        return self.__all_output

    def get_lines(self):
        return self.__all_lines



from pearlib.apply_nonlinearity import apply_nonlinearity
from pearlib.sigmoid import *


def calculate_all_output(layers, res):
    """Calculates the  output of the layers for all the possible inputs of the network in size.

        layers ~ ([layer_i]) -- Ordered list of the network's layers. (Weight matrices and bias vectors)
        res -- resolution of the plane

        The output is calculated in the following way:
            h_0 = apply_nonlinearity(W_0*x+b_0)
            h_i = apply_nonlinearity(W_i*h_{i-1}+b_i)   for i = 1..layer_idx
                for the last layer: sigmoid(W_i*h_{i-1}+b_i)

        Returns the output(s) of all the layers in a list.
            the first element of this list is also a list, containing the outputs of the first layer, etc
    """

    # Generate input field - it is always 2 dimensional, [-1,1]x[-1,1]
    x = np.linspace(-1, 1, res[0])
    y = np.linspace(-1, 1, res[1])

    # Initialize h_new for the first layer
    h_new = np.zeros((layers[0].n_input, len(x) * len(y)))
    for i in range(0, len(x)):
        for j in range(0, len(y)):
            h_new[:, i * len(y) + j] = [x[i], y[j]]

    # Array containing the outputs
    all_output = []
    # For each layer
    for layer_i in range(0, len(layers)):
        # Save the output of the previous layer/input of the layer to h_old
        h_old = h_new

        # Calculate the current layer's output
        del h_new
        h_new = layers[layer_i].calculate_layer_output_mat(h_old)

        # If it is not the output layer: apply the nonlinearity
        if layer_i != len(layers) - 1:
            apply_nonlinearity(h_new)
        # If it is the last layer: apply sigmoid
        else:
            h_new = sigmoid(h_new)

        # Clear the input (it will be overwritten in the next iteration)
        del h_old

        all_output.append(np.reshape(h_new.transpose(), [len(x), len(y), h_new.shape[0]], 'C'))

    # Return all the outputs
    return all_output



def get_color_for_layeridx(layeridx):
    """For an index it gives back an associated color (3 channel, each between 0 and 1)"""

    # For the first six layer: base colors
    if layeridx < 6:
        base6_colors = [[1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 0], [1, 0, 1], [0, 1, 1]]
        return base6_colors[layeridx]
    # From the 7th index: calculate a unique color
    else:
        # increment in red component
        ro = 161
        # increment in blue component
        bo = 111
        # increment in green component
        go = 67
        # maximum color
        nc = 256.0
        if layeridx % 3 == 0:
            return [((layeridx - 5) * ro % nc) / nc, ((layeridx - 6) * bo % nc) / nc, ((layeridx - 6) * go % nc) / nc]
        elif layeridx % 3 == 1:
            return [((layeridx - 6) * ro % nc) / nc, ((layeridx - 5) * bo % nc) / nc, ((layeridx - 6) * go % nc) / nc]
        else:
            return [((layeridx - 6) * ro % nc) / nc, ((layeridx - 6) * bo % nc) / nc, ((layeridx - 5) * go % nc) / nc]



#####################################################
#####################################################
## PEAR: EXAMPLE FILE TO VISUALIZE NEURAL NETWORKS ##
#####################################################
#####################################################

import numpy as np
import matplotlib.pyplot as plt
from pearlib.network import Network
from pearlib.create_image import create_image
from pearlib.save_image import save_image
from pearlib.calculate_output import calculate_output
from pearlib.get_all_lines import get_all_lines
from pearlib.create_image_from_lines import create_image_from_lines

# Path of the network to load.
network_name = "./3000_parameters/lena.jpg"
# Initializing Pear's inner representation for networks
my_network = Network(network_name)

resolution = [301, 301]

# Extracting layers and computing all the outputs for the given resolution
layers = my_network.get_layers()
lines = get_all_lines(my_network)

fig, ax = plt.subplots()


###############################################################
# Case 1: visualizing the last layer on a blank background    #
###############################################################
image_from_scratch = create_image(my_network, [len(layers) - 2], resolution, None, False)

ax.imshow(image_from_scratch, cmap = 'gray' ,interpolation = 'nearest', origin = 'upper')


###############################################################
# Case 2: visualizing the last layer on approximated image    #
###############################################################
# This time, we will not recompute everything from scratch:
# we already have the lines, we just need to draw them
image_on_approximation = create_image_from_lines(lines, [len(layers) - 2], resolution, None, my_network, True)

fig, ax = plt.subplots()
ax.imshow(image_on_approximation, cmap = 'gray' ,interpolation = 'nearest', origin = 'upper')


###############################################################
# Case 3: visualizing the last layer on a background image    #
###############################################################
lena = plt.imread('./lena.jpg')
lena = lena / 255 # Normalizing

image_with_background = create_image(my_network, [len(layers) - 2], None, lena, False)
fig, ax = plt.subplots()
ax.imshow(image_with_background, cmap = 'gray' ,interpolation = 'nearest', origin = 'upper')


plt.show()


##############################################################
# Saving a created image to the current folder               #
##############################################################
save_image('./test_lena.png', image_with_background)



import numpy as np
from pearlib.draw_line import draw_line
from pearlib.layer import *
from pearlib.sigmoid import *


def create_image_from_lines(line_list, layer_indices, res = None, img = None, network = None , plot_output = False):
    """Generates the image with the boundary lines.
        -- line_list: contains all the lines from all the layers (list of lists)
        -- layer_indices: indicates which layers we would like to draw
        -- res: resolution of the plane ([n_col, n_row])
        -- img: optional. If given, the lines will be plotted on that image. If given, its resolution is used.
        -- network: The network object. Only used when plot output is true
        -- plot_output: optional. If true, the network should be given as well.
            If true, the background of the image will be the output of the neuron in the last layer
    """

    # Tests for the pair-inputs
    # If neither img nor res is given -> error, we do not know the resolution
    if img is None and res is None:
        raise ValueError("No resolution specified")
    # If the network is not given -> error, we can not calculate the output
    if network is None and plot_output:
        raise ValueError("No network is given to calculate the output image.")

    # If we should plot the network's output, get it
    if plot_output:
        # If in the last layer we truly have only 1 neuron
        layers = network.get_layers()
        if layers[(len(layers) - 1)].n_neurons == 1:
            # Get the output of the last layer
            all_output = network.get_all_output()
            # Copy the output of the last layer to be the background
            img = np.array(all_output[len(all_output) - 1], copy = True)
            # Expand the image to be colored
            img = np.tile(img, (1, 1, 3))
        # If in the last layer we have more than one neuron
        else:
            raise ValueError("Error in create_image_from_lines: The number of neurons in the last layer should be one.")

    # If plot_output is False
    else:
        # If the image is not given -> create it with the given resolution (each element is 1 (white))
        if img is None:
            x = np.linspace(-1, 1, res[0])
            y = np.linspace(-1, 1, res[1])
            img = np.ones((len(x), len(y), 3))

        # If the image is given
        else:
            # If the image is grayscale (has only 1 channel)
            if len(img.shape) == 2 or img.shape[2] == 1:
                img = np.transpose(img)
                img = img[:, :, np.newaxis]
                img = np.tile(img, (1, 1, 3))
            # If the image is colored (has 3 channels)
            elif img.shape[2] == 3:
                img = np.transpose(img, (1, 0, 2))
            # If the image has 4 channels, it is an RGBA image -> drop the fourth channel       
            elif img.shape[2] == 4:       
                img = img[:, :, :3]       
                img = np.transpose(img, (1, 0, 2))
            else:
                raise ValueError("The given image is neither a grayscale nor a colored RGB image\n"
                                 "(Does not have 1 nor 3 channels)")

    # For all the layers in layer_indices, draw the boundary lines
    for i in range(len(layer_indices)):
        # If it is not the last layer
        if layer_indices[i] < len(line_list):
            # for all the lines in the given layer
            for j in range(len(line_list[layer_indices[i]])):
                # draw the line to the image
                draw_line(img, line_list[layer_indices[i]][j])

    img = np.transpose(img, (1, 0, 2))
    return img



import numpy as np
from pearlib.line import Line


def draw_line(img, line):
    """Draw the line given in 'line' on 'img'."""
    # For each pixel indicated in line we set the img to its color.
    for i in range(0, line.points.shape[1]):
        idx = line.points[:, i]
        img[idx[0], idx[1], :] = line.color



__all__ = ["apply_nonlinearity", "calculate_all_output", "calculate_output", "create_image", "create_image_from_lines", "draw_line", "get_all_lines", "get_color_for_layeridx", "get_lines_from_layer_output", "layer", "line", "network", "reader", "save_image", "sigmoid"]


from setuptools import setup

def readme():
	with open('README.rst') as f:
		return f.read()

setup(name = 'pearlib',
      version = '0.9',
      description = 'The library that makes neural networks appear',
      long_description = readme(),
      classifiers = [
      'Development Status :: 3 - Alpha',
      'License :: OSI Approved :: BSD License',
      'Natural Language :: English',
      'Programming Language :: Python :: 3.6',
      'Topic :: Scientific/Engineering :: Visualization',
      'Framework :: IDLE'],
      keywords = 'tensorflow neural networks visualization',
      python_requires = '>=3',
      url = 'https://github.com/danimano/TRP',
      author = 'Candice Bentéjac, Anna Csörgő and Dániel Hajtó',
      author_email = 'candice.bentejac@etu.u-bordeaux.fr',
      license = 'BSD-3-Clause',
      packages = ['pearlib'],
      install_requires = ['numpy', 'matplotlib', 'Pillow', 'tensorflow', 'webcolors'],
      zip_safe = False)




import numpy as np
from pearlib.layer import Layer
from pearlib.apply_nonlinearity import apply_nonlinearity
from pearlib.sigmoid import *


def calculate_output(layers, res, layer_idx):
    """Calculates the  output of the given layer for all the possible inputs of the network in size.

        layers -- Ordered list of layers.
                               (Weight matrices and bias vectors of the network.)
        res -- resolution of the plane
        layer_idx -- index of the layer, for which we want to determine the output

        The output is calculated in the following way:
            h_0 = apply_nonlinearity(W_0*x+b_0)
            h_i = apply_nonlinearity(W_i*h_{i-1}+b_i)   for i = 1..layer_idx
                for the last layer: sigmoid(W_i*h_{i-1}+b_i)

        Returns the output(s) of the queried layer in a 3D array, where each plane represents the output of one neuron.
    """

    if layer_idx >= len(layers):
        raise ValueError("layer_idx exceeds the number of layers in the network")

    # Generate input field - it is always 2 dimensional, [-1,1]x[-1,1]
    x = np.linspace(-1, 1, res[0])
    y = np.linspace(-1, 1, res[1])

    # Initialize h_new for the first layer
    h_new = np.zeros((layers[0].n_input, len(x) * len(y)))
    for i in range(0, len(x)):
        for j in range(0, len(y)):
            h_new[:, i * len(y) + j] = [x[i], y[j]]

    # For each layer 0..layer_idx
    for layer_i in range(0, layer_idx + 1):
        # Save the output of the previous layer/input of the current layer to h_old
        h_old = h_new

        # Calculate the layer's output
        del h_new
        h_new = layers[layer_i].calculate_layer_output_mat(h_old)

        # If it is not the output layer: apply the nonlinearity
        if layer_i != len(layers) - 1:
            apply_nonlinearity(h_new)
        # If it is the last layer apply sigmoid
        else:
            h_new = sigmoid(h_new)

        # Clear the input (it will be overwritten in the next iteration)
        del h_old

    # Convert the output of the layer in the above defined format
    h_new = np.reshape(h_new.transpose(), [len(x), len(y), h_new.shape[0]], 'C')

    # Return the output of the layer_idxth layer
    return h_new



def apply_nonlinearity(h):
    """Applies ReLU on each element of h.
        h -- nD array
             ReLU:
                0 if h <= 0
                h if h > 0
        Thresholds the negative values in the input array to 0.
    """

    low_values_flags = h < 0  # Where values are low
    h[low_values_flags] = 0  # All low values set to 0



class Line:
    """Defines a boundary line object"""
    # Each column contains th ecoordinates of a point pf the line
    points = []
    # Color of the line (unique for each layer)
    color = [1, 1, 1]

    def __init__(self, _points= [], _color = [1, 1, 1]):
        self.points = _points
        self.color = _color
